// This module creates the block structure.
// The FSM checks if there is an event ready for transmission.
// Must register all #s and flags at the beginning and then
// give a unique pop for ctrl_num_fifos and pointer_fifo.

`include "define.v"
module DaqFormatter
( input clk,  rst, TestMode, DInError, 
        OW, BZ, LW, DLL_lock, StRq, TTC_Ready, EventBufferEmpty,
        EmEv, RL, TrigMode, CompactNorm, CompactCalib, EvtRdy4Tx,
  input [1:0] UnlockCnt,
  input [3:0] Fifo1Status, Fifo2Status, Fifo3Status, Fifo4Status, 
              Fifo5Status, Fifo6Status, Fifo7Status, Fifo8Status,   
  input [4:0] NS,  NPS,
  input [7:0] PipeLength,  
  input [8:1] FifoEmpty, FifoFull, // Conditions of the 8 Input FIFOs.
  input [7:0] TPwordCount,
  input [8:0] DAQwordCount,
  input [`MAIN_DATA_WIDTH -1:0]  data_in, // from the event_buffer
  input [`EVN_WIDTH - 1:0] EvN, CTrEvN,      
  input [`BCN_WIDTH -1:0]  bcn,         
         Extra1Info, Extra2Info, Extra3Info, Extra4Info,
         Extra5Info, Extra6Info, Extra7Info, Extra8Info,
  input [`OrN_WIDTH -1:0] orn,       
  input [`HCN_WIDTH -1:0] htr_card_no, 
  output reg EventBufferPop, RdyAck,
  output reg FifosPop, // pops all fifos but event_buffer
  output reg [1:0] StatusBits, //[S1,S0]
  output reg [4:0] DAQ_STATE,  // output for debugging only  
  output reg [9:0] WordCount,  // output for debugging only  
  output reg [`MAIN_DATA_WIDTH-1:0] DataOut      );   

parameter  
   IDLE           =   0,
   HEADER         =   1,             
   HEADER_2       =   2,
   HEADER_3       =   3,
   HEADER_4       =   4, 
   HEADER_5       =   5,
   HEADER_6       =   6, 
   HEADER_7       =   7, 
   HEADER_8       =   8, 
   DATA           =   9, 
   PARITY_COMPACT =  24,
   PARITY_RICH    =  25,
   BZERO1         =  11,  
   BZERO2         =  12,  
   BZERO3         =  13,  
   BZERO4         =  14,  
   BZERO5         =  15,  
   BZERO6         =  16,  
   BZERO7         =  17,  
   BZERO8         =  18,  
   PRETRAILER3    =  19,
   PRETRAILER2    =  20,
   PRETRAILER1    =  21,
   TRAILER        =  22;

// ***********  INTERNAL SIGNALS   *********************** 
 reg  [7:0] LatchedTPwordCount;
 reg  [8:0] LatchedDAQwordCount;
 reg  [`EVN_WIDTH - 1:0] latched_EvN; // latch retains the value 
 reg  [`BCN_WIDTH -1:0]  latched_bcn;         
 reg  [`OrN_WIDTH -1:0]  latched_orn;       
 reg  latched_EmEV, latched_RL, latched_TrigMode, OD, rich_or_compact,  
      latched_DL, latched_TR, latched_LW, old_LW, LE;   
   
always@(posedge clk)
    begin   // always
    if (rst)   begin  // rst
      DAQ_STATE         <= IDLE;    StatusBits       <= 'd0;
      DataOut           <= 'd0;     RdyAck           <= 0;
      EventBufferPop    <= 0;       FifosPop         <= 0;
      WordCount         <= 0;       LE <= 0;
    end  // rst
   else begin
      // the following bits are de-asserted on state HEADER_4:
      if (~DLL_lock)  latched_DL  <= 0; // Latch a DLL problem 
      if (~TTC_Ready) latched_TR  <= 0; // Latch a TTCrx problem 
      if (DInError)   OD  <= 1; // Latch an Optical Data error  
      if (LW)         latched_LW  <= 1;
      case(DAQ_STATE)
      IDLE:    begin 
        StatusBits       <= 0;
        DataOut          <= 0;
        if(EvtRdy4Tx)  // there is a complete event in the EventBuffer...
          begin        //    so start the transmission af a block  
          DAQ_STATE        <= HEADER;
          WordCount        <= 'd3; // start from 3 as it won't increment at the last 3 states
          RdyAck           <= 1;
          // latch tags as they might change later (upon new fifo_pops)
          latched_RL          <= RL;           latched_EmEV        <= EmEv;
          latched_bcn         <= bcn;          latched_orn         <= orn;             
          latched_TrigMode    <= TrigMode;
          LatchedDAQwordCount <= DAQwordCount; LatchedTPwordCount <= TPwordCount;
          if (TrigMode) latched_EvN  <= CTrEvN; 
          else          latched_EvN  <= EvN; 
          end    //  EvRdy4Transm
      end  // IDLE
      HEADER:   begin 
         WordCount         <= WordCount + 1;
         StatusBits        <= 'b11;         
         DataOut[15]       <= StRq;
         DataOut[14:8]     <= 'd0;
         DataOut[7:0]      <= latched_EvN[7:0];
         DAQ_STATE         <= HEADER_2;
         FifosPop          <= 1;
         RdyAck            <= 0;
         if ((latched_TrigMode & CompactCalib) | 
             (~latched_TrigMode & CompactNorm)   ) 
                rich_or_compact <= 1'b1;                  
         else   rich_or_compact <= 1'b0;  // for later use
      end                                
      HEADER_2: begin
         WordCount         <= WordCount + 1;
         StatusBits        <= 'b10;
         DataOut           <= latched_EvN[23:8];
         DAQ_STATE         <= HEADER_3;
         FifosPop          <= 0;
      end
      HEADER_3: begin
         WordCount         <= WordCount + 1;
         StatusBits        <= 'b10;
         DataOut[0]        <= OW;
         DataOut[1]        <= BZ;
         DataOut[2]        <= latched_EmEV;  
         DataOut[3]        <= latched_RL;
         DataOut[4]        <= LE;
         DataOut[5]        <= latched_LW;
         DataOut[6]        <= OD;  
         DataOut[7]        <= ~(latched_DL & latched_TR);  
         DataOut[11:8]     <= 0;  
         DataOut[12]       <= TestMode;
         DataOut[13]       <= 0;  // -> normal firmware (non histograming)
         DataOut[14]       <= latched_TrigMode;  
         DataOut[15]       <= 1'b1;   
         DAQ_STATE         <= HEADER_4;
      end
      HEADER_4:  begin  
         WordCount         <= WordCount + 1;
         StatusBits        <= 'b10;
         DataOut[9:0]      <= htr_card_no; 
         DataOut[15:10]    <= latched_orn[5:0];
         DAQ_STATE         <= HEADER_5;
         latched_RL        <= 0;    OD          <= 0;    
         latched_DL        <= 1;    latched_TR  <= 1;
         latched_LW        <= 0;    old_LW      <= latched_LW;
         if (latched_LW & old_LW) LE <= 1;
      end
      HEADER_5:   begin
         WordCount           <= WordCount + 1;  // here WordCount <= 8  always.
         StatusBits          <= 'b10;
         DataOut[11:0]       <= latched_bcn;
         DataOut[15:12]      <= `FORMAT_VERSION;
         if(~latched_EmEV)   begin
           DAQ_STATE        <=   HEADER_6; 
         end  // if
         else    begin  // this is an Empty Event
             DAQ_STATE <= PRETRAILER2;
         end  // else
      end            
      HEADER_6:   begin
         WordCount           <= WordCount + 1;  // if FSM arrives here: WordCount <= 9.
         StatusBits          <= 'b10;
         DataOut[0]          <=  TTC_Ready;   
         DataOut[2:1]        <=  UnlockCnt;   
         DataOut[7:3]        <=  NPS;   
         DataOut[15:8]       <=  LatchedTPwordCount;   
         DAQ_STATE <= HEADER_7;
      end   // HEADER_6
      HEADER_7:   begin
         WordCount           <= WordCount + 1;  
         StatusBits          <= 'b10;
         DataOut[12:0]       <= `FPGAversionN;   
         DataOut[15:13]      <= `FPGAversionType;   
         EventBufferPop      <= 1;  // prepare the 2nd data after 2 ticks !
         DAQ_STATE <= HEADER_8;
      end   // HEADER_7
      HEADER_8:   begin
         WordCount           <= WordCount + 1;  
         StatusBits          <= 'b10;
         DataOut[7:0]        <= PipeLength;   
         DataOut[15:8]       <= 0;   
         EventBufferPop      <=  1;   
         DAQ_STATE <= DATA;
      end   // HEADER_8

      // The first execution of DATA will latch the first DataOut, that was 
      // automatically ready because of the flow-thru mode of event-buffer.   
      DATA:   begin       // decimal 9
         WordCount           <= WordCount + 1;
         StatusBits          <= 'b10;
         DataOut             <= data_in;            
         if (EventBufferEmpty)  begin
             EventBufferPop    <= 0;
             case ({WordCount[0], rich_or_compact})
                  2'b00  :  DAQ_STATE   <= BZERO1;
                  2'b01  :  DAQ_STATE   <= PRETRAILER3;
                  2'b10  :  DAQ_STATE   <= PARITY_RICH;
                  2'b11  :  DAQ_STATE   <= PARITY_COMPACT;
            endcase 
          end // if
      end   // DATA      
      PARITY_COMPACT: begin   // decimal 24 = hex 18
         WordCount          <= WordCount + 1;
         StatusBits         <= 'b10;
         DataOut            <= 16'hFFFF;            
         DAQ_STATE          <= PRETRAILER3;
      end
      PARITY_RICH: begin      // decimal 25 = hex 19
         WordCount          <= WordCount + 1;
         StatusBits         <= 'b10;
         DataOut            <= 16'hFFFF;            
         DAQ_STATE          <= BZERO1;
      end
      BZERO1:       begin     // decimal 11 = hex B
          WordCount          <= WordCount + 1;  
          StatusBits         <= 'b10;          
          DataOut[11:0]      <= Extra1Info;
          DataOut[15:12]     <= {FifoEmpty[1], FifoFull[1], Fifo1Status[1:0]};
          DAQ_STATE          <= BZERO2;
      end            
      BZERO2:       begin
          WordCount          <= WordCount + 1;
          StatusBits         <= 'b10;
          DataOut[11:0]      <= Extra2Info;
          DataOut[15:12]     <= {FifoEmpty[2], FifoFull[2], Fifo2Status[1:0]};
          DAQ_STATE          <= BZERO3;   
      end            
      BZERO3:       begin
          WordCount          <= WordCount + 1;
          StatusBits         <= 'b10;
          DataOut            <= Extra3Info;
          DataOut[15:12]     <= {FifoEmpty[3], FifoFull[3], Fifo3Status[1:0]};
          DAQ_STATE          <= BZERO4;   
      end            
      BZERO4:       begin
          WordCount          <= WordCount + 1;
          StatusBits         <= 'b10;
          DataOut[11:0]      <= Extra4Info;
          DataOut[15:12]     <= {FifoEmpty[4], FifoFull[4], Fifo4Status[1:0]};
          DAQ_STATE          <= BZERO5;   
      end            
      BZERO5:       begin
          WordCount          <= WordCount + 1;
          StatusBits         <= 'b10;
          DataOut[11:0]      <= Extra5Info;
          DataOut[15:12]     <= {FifoEmpty[5], FifoFull[5], Fifo5Status[1:0]};
          DAQ_STATE          <= BZERO6;   
      end            
      BZERO6:       begin
          WordCount          <= WordCount + 1;
          StatusBits         <= 'b10;
          DataOut[11:0]      <= Extra6Info;
          DataOut[15:12]     <= {FifoEmpty[6], FifoFull[6], Fifo6Status[1:0]};
          DAQ_STATE          <= BZERO7;   
      end            
      BZERO7:       begin
          WordCount          <= WordCount + 1;
          StatusBits         <= 'b10;
          DataOut[11:0]      <= Extra7Info;
          DataOut[15:12]     <= {FifoEmpty[7], FifoFull[7], Fifo7Status[1:0]};
          DAQ_STATE          <= BZERO8;   
      end            
      BZERO8:       begin
          WordCount          <= WordCount + 1;
          StatusBits         <= 'b10;
          DataOut[11:0]      <= Extra8Info;
          DataOut[15:12]     <= {FifoEmpty[8], FifoFull[8], Fifo8Status[1:0]};
          DAQ_STATE          <= PRETRAILER3;   
      end            
            
      PRETRAILER3:  begin   
          WordCount          <= WordCount + 1;
          StatusBits         <= 'b10;
          DataOut[15:11]     <= NS;
          DataOut[10:0]      <= {2'b0, LatchedDAQwordCount}; 
          DAQ_STATE          <= PRETRAILER2;   
      end            
      PRETRAILER2:  begin
          StatusBits         <= 'b10;
          DataOut            <= WordCount;
          DAQ_STATE          <= PRETRAILER1;   
      end            
      PRETRAILER1:   begin   
          StatusBits        <=  'b10;
          DataOut           <=  'd0;
          DAQ_STATE         <=  TRAILER; 
          EventBufferPop    <=  0;
      end            
      TRAILER:       begin
          StatusBits       <= 'b01;
          DataOut          <= {latched_EvN[7:0], 8'b0};
          DAQ_STATE        <= IDLE;
      end   // TRAILER         
      default: DAQ_STATE <= IDLE;
      endcase   
      end
      end  // always
endmodule